# This file is part of GridCal.
#
# GridCal is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GridCal is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GridCal.  If not, see <http://www.gnu.org/licenses/>.

import pandas as pd
import time
from PySide2.QtCore import QThread, Signal

from GridCal.Engine.basic_structures import Logger
from GridCal.Engine.basic_structures import TimeGrouping, get_time_groups
from GridCal.Engine.Core.multi_circuit import MultiCircuit
from GridCal.Engine.Simulations.PowerFlow.power_flow_worker import SolverType
from GridCal.Engine.Simulations.OPF.opf_driver import OptimalPowerFlowOptions
from GridCal.Engine.Simulations.OPF.dc_opf_ts import OpfDcTimeSeries
from GridCal.Engine.Simulations.OPF.ac_opf_ts import OpfAcTimeSeries
from GridCal.Engine.Simulations.OPF.simple_dispatch_ts import OpfSimpleTimeSeries
from GridCal.Engine.Core.time_series_opf_data import compile_opf_time_circuit
from GridCal.Engine.Simulations.OPF.opf_ts_results import OptimalPowerFlowTimeSeriesResults


class OptimalPowerFlowTimeSeries(QThread):
    progress_signal = Signal(float)
    progress_text = Signal(str)
    done_signal = Signal()
    name = 'Optimal power flow time series'

    def __init__(self, grid: MultiCircuit, options: OptimalPowerFlowOptions, start_=0, end_=None):
        """
        PowerFlowDriver class constructor
        @param grid: MultiCircuit Object
        @param options: OPF options
        """
        QThread.__init__(self)

        # Grid to run a power flow in
        self.grid = grid

        # Options to use
        self.options = options

        # power flow options
        self.pf_options = options.power_flow_options

        # compile the circuit into a numerical equivalent for this simulation
        self.numerical_circuit = compile_opf_time_circuit(circuit=self.grid,
                                                          apply_temperature=self.pf_options.apply_temperature_correction,
                                                          branch_tolerance_mode=self.pf_options.branch_impedance_tolerance_mode)

        # OPF results
        self.results = OptimalPowerFlowTimeSeriesResults(bus_names=self.numerical_circuit.bus_names,
                                                         branch_names=self.numerical_circuit.branch_names,
                                                         load_names=self.numerical_circuit.load_names,
                                                         generator_names=self.numerical_circuit.generator_names,
                                                         battery_names=self.numerical_circuit.battery_names,
                                                         n=self.grid.get_bus_number(),
                                                         m=self.grid.get_branch_number(),
                                                         nt=len(self.grid.time_profile),
                                                         ngen=len(self.grid.get_generators()),
                                                         nbat=len(self.grid.get_batteries()),
                                                         nload=len(self.grid.get_loads()),
                                                         time=self.grid.time_profile,
                                                         bus_types=self.numerical_circuit.bus_types)

        self.start_ = start_

        if end_ is not None:
            self.end_ = end_
        else:
            self.end_ = len(self.grid.time_profile)

        self.logger = Logger()

        # set cancel state
        self.__cancel__ = False

        self.all_solved = True

        self.elapsed = 0.0

    def reset_results(self):
        """
        Clears the results
        """
        # reinitialize
        self.results = OptimalPowerFlowTimeSeriesResults(bus_names=self.numerical_circuit.bus_names,
                                                         branch_names=self.numerical_circuit.branch_names,
                                                         load_names=self.numerical_circuit.load_names,
                                                         generator_names=self.numerical_circuit.generator_names,
                                                         battery_names=self.numerical_circuit.battery_names,
                                                         n=self.grid.get_bus_number(),
                                                         m=self.grid.get_branch_number(),
                                                         nt=len(self.grid.time_profile),
                                                         ngen=len(self.grid.get_generators()),
                                                         nbat=len(self.grid.get_batteries()),
                                                         nload=len(self.grid.get_loads()),
                                                         time=self.grid.time_profile,
                                                         bus_types=self.numerical_circuit.bus_types)

    def get_steps(self):
        """
        Get time steps list of strings
        """
        return [l.strftime('%d-%m-%Y %H:%M') for l in pd.to_datetime(self.grid.time_profile)]

    def opf(self, start_, end_, remote=False, batteries_energy_0=None):
        """
        Run a power flow for every circuit
        :param start_: start index
        :param end_: end index
        :param remote: is this function being called from the time series?
        :param batteries_energy_0: initial state of the batteries, if None the default values are taken
        :return: OptimalPowerFlowResults object
        """

        if not remote:
            self.progress_signal.emit(0.0)
            self.progress_text.emit('Formulating problem...')

        if self.options.solver == SolverType.DC_OPF:

            # DC optimal power flow
            problem = OpfDcTimeSeries(numerical_circuit=self.numerical_circuit,
                                      start_idx=start_,
                                      end_idx=end_,
                                      solver=self.options.mip_solver,
                                      batteries_energy_0=batteries_energy_0)

        elif self.options.solver == SolverType.AC_OPF:

            # AC optimal power flow
            problem = OpfAcTimeSeries(numerical_circuit=self.numerical_circuit,
                                      start_idx=start_,
                                      end_idx=end_,
                                      solver=self.options.mip_solver,
                                      batteries_energy_0=batteries_energy_0)

        elif self.options.solver == SolverType.Simple_OPF:

            # AC optimal power flow
            problem = OpfSimpleTimeSeries(numerical_circuit=self.numerical_circuit,
                                          start_idx=start_, end_idx=end_,
                                          solver=self.options.mip_solver,
                                          batteries_energy_0=batteries_energy_0,
                                          text_prog=self.progress_text.emit,
                                          prog_func=self.progress_signal.emit)

        else:
            self.logger.append('Solver not supported in this mode: ' + str(self.options.solver))
            return

        if not remote:
            self.progress_signal.emit(0.0)
            self.progress_text.emit('Running all in an external solver, this may take a while...')

        # solve the problem
        status = problem.solve()
        print("Status:", status)

        a = start_
        b = end_
        self.results.voltage[a:b, :] = problem.get_voltage()
        self.results.load_shedding[a:b, :] = problem.get_load_shedding()
        self.results.battery_power[a:b, :] = problem.get_battery_power()
        self.results.battery_energy[a:b, :] = problem.get_battery_energy()
        self.results.generator_power[a:b, :] = problem.get_generator_power()
        self.results.Sbranch[a:b, :] = problem.get_branch_power()
        self.results.overloads[a:b, :] = problem.get_overloads()
        self.results.loading[a:b, :] = problem.get_loading()
        self.results.shadow_prices[a:b, :] = problem.get_shadow_prices()

        return self.results

    def opf_by_groups(self):
        """
        Run the OPF by groups
        """

        self.progress_signal.emit(0.0)
        self.progress_text.emit('Making groups...')

        # get the partition points of the time series
        groups = get_time_groups(t_array=self.grid.time_profile, grouping=self.options.grouping)

        n = len(groups)
        i = 1
        energy_0 = None
        while i < n and not self.__cancel__:

            start_ = groups[i - 1]
            end_ = groups[i]

            print(start_, ':', end_, ' [', end_ - start_, ']')

            if start_ >= self.start_ and end_ <= self.end_:

                # run an opf for the group interval only if the group is within the start:end boundaries
                self.opf(start_=start_, end_=end_ + 1, remote=True, batteries_energy_0=energy_0)

            energy_0 = self.results.battery_energy[end_ - 1, :]

            self.progress_text.emit('Running OPF for the time group ' + str(i) + ' in external solver...')
            progress = ((start_ - self.start_ + 1) / (self.end_ - self.start_)) * 100
            self.progress_signal.emit(progress)

            i += 1

    def run(self):
        """

        :return:
        """

        start = time.time()

        if self.options.grouping == TimeGrouping.NoGrouping:
            self.opf(start_=self.start_, end_=self.end_)
        else:
            self.opf_by_groups()

        end = time.time()
        self.elapsed = end - start

        self.progress_signal.emit(0.0)
        self.progress_text.emit('Done!')
        self.done_signal.emit()

    def cancel(self):
        self.__cancel__ = True
        self.progress_signal.emit(0.0)
        self.progress_text.emit('Cancelled!')
        self.done_signal.emit()



"""
.. module:: mpiserve
   :synopsis: MPI-based controller server and workers for POAP.
.. moduleauthor:: David Bindel <bindel@cornell.edu>
"""

# NB: Must do mpirun with a working mpi4py install.
#     See https://groups.google.com/forum/#!topic/mpi4py/ULMq-bC1oQA

try:
    import Queue
except ImportError:
    import queue as Queue

import logging

from threading import Thread
from mpi4py import MPI
from poap.controller import Controller

# Get module-level logger
logger = logging.getLogger(__name__)

# Get MPI communicator and rank
comm = MPI.COMM_WORLD
rank = comm.Get_rank()
nproc = comm.Get_size()


class MPIController(Controller):
    """MPI controller.

    The MPI controller *must* run at rank 0.

    The server sends messages of the form
        ('eval', record_id, args, extra_args)
        ('eval', record_id, args)
        ('kill', record_id)
        ('terminate')
    The default messages received are
        ('update_dict', record_id, dict)
        ('running', record_id)
        ('kill', record_id)
        ('cancel', record_id)
        ('complete', record_id, value)
    """

    def __init__(self, strategy=None):
        "Initialize the controller."
        Controller.__init__(self)
        self._workers = [w for w in range(1, nproc)]
        self._recids = {}
        self._needs_ping = False
        self.strategy = strategy
        self.add_term_callback(self._send_shutdown)

    def ping(self):
        "Ping controller to check in with strategies"
        logger.debug("Ping MPIController to check strategies")
        self._needs_ping = False

    def can_work(self):
        "Return whether we can currently perform work."
        return len(self._workers) > 0

    def _handle_message(self):
        """Handle received messages.

        Receive record update messages of the form
            ('action', record_id, params)
        where 'action' is the name of an EvalRecord method and params is
        the list of parameters.  The record_id should be recorded in the
        hub's records table (which happens whenever it is referenced in
        a message sent to a worker).

        On a message indicating that the worker is done with the record,
        we add the worker that sent the message back to the free pool.
        """
        logger.debug("Handle incoming message")
        s = MPI.Status()
        data = comm.recv(status=s, source=MPI.ANY_SOURCE, tag=0)
        logger.debug("Received message: %s", data)
        mname = data[0]
        record = self._recids[data[1]]
        method = getattr(record, mname)
        method(*data[2:])
        if mname == 'complete' or mname == 'cancel' or mname == 'kill':
            logger.debug("Re-queueing worker")
            self._workers.append(s.source)
        self.ping()

    def _submit_work(self, proposal):
        "Create new record and send to worker"
        worker = self._workers.pop()
        record = self.new_feval(proposal.args)
        record.worker = worker
        proposal.record = record
        self._recids[id(record)] = record
        proposal.accept()
        logger.debug("Dispatch eval request to %d", worker)
        if record.extra_args is None:
            m = ('eval', id(record), record.params)
        else:
            m = ('eval', id(record), record.params, record.extra_args)
        comm.send(m, dest=worker, tag=0)

    def _kill_work(self, record):
        "Send a kill request to a worker"
        worker = record.worker
        logger.debug("Dispatch kill request to %d", worker)
        comm.send(('kill', id(record)), dest=worker, tag=0)

    def _send_shutdown(self):
        "Send shutdown requests to all workers"
        for worker in range(1, nproc):
            comm.send(('terminate',), dest=worker, tag=0)

    def _run(self, merit=None, filterp=None):
        "Run the optimization and return the best value."
        while True:
            if comm.Iprobe():
                self._handle_message()
            proposal = None
            if not self._needs_ping:
                proposal = self.strategy.propose_action()
            if not proposal:
                self._handle_message()
            elif proposal.action == 'terminate':
                logger.debug("Accept terminate proposal")
                proposal.accept()
                return self.best_point(merit=merit, filter=filterp)
            elif proposal.action == 'eval' and self.can_work():
                logger.debug("Accept eval proposal")
                self._submit_work(proposal)
            elif proposal.action == 'kill' and not proposal.args[0].is_done:
                logger.debug("Accept kill proposal")
                record = proposal.args[0]
                proposal.accept()
                self._kill_work(record)
            else:
                logger.debug("Reject proposal")
                proposal.reject()
                self._needs_ping = True

    def run(self, merit=None, filterp=None):
        """Run the optimization and return the best value.

        Args:
            merit: Function to minimize (default is r.value)
            filterp: Predicate to use for filtering candidates

        Returns:
            Record minimizing merit() and satisfying filterp();
            or None if nothing satisfies the filter
        """
        try:
            return self._run(merit=merit, filterp=filterp)
        finally:
            self.call_term_callbacks()


class MPIWorker(object):
    """MPI worker process.

    An MPI worker mostly monitors messages from the master.
    It also monitors a local queue in case there is a separate
    computational process or thread generating results.  The
    main routine for the evaluator is eval.
    """

    def __init__(self):
        self._running = False
        self._outbox = Queue.Queue()
        self._eval_thread = None
        self._eval_killed = False
        self.daemonize = False

    def eval(self, record_id, args, extra_args=None):
        "Actually do the function evaluation (separate thread)"
        logger.error("Call to base MPIWorker.eval(%d, %s, %s)",
                     record_id, args, extra_args)

    def _eval_wrap(self, record_id, args, extra_args=None):
        "Wrapper for eval request (handle evaluation crash)"
        try:
            if extra_args is None:
                self.eval(record_id, args)
            else:
                self.eval(record_id, args, extra_args)
        except Exception:
            logger.error("Function evaluation failed")
            self.finish_cancel(record_id)

    def on_eval(self, record_id, args, extra_args=None):
        "Handle eval request."
        logger.debug("In MPIWorker.on_eval")
        self._eval_thread = Thread(target=self._eval_wrap,
                                   args=(record_id, args, extra_args))
        self._eval_thread.daemon = self.daemonize
        self._eval_killed = False
        self._eval_thread.start()
        logger.debug("Returning after worker thread start")

    def on_kill(self, record_id):
        "Handle kill request."
        logger.debug("In MPIWorker.on_kill(%d)", record_id)
        self._eval_killed = True

    def on_terminate(self):
        "Handle termination request."
        logger.debug("In MPIWorker.on_terminate")
        self._running = False

    def send(self, *args):
        "Queue message to process 0 (where the controller lives)."
        logger.debug("Queue outgoing message %s", args)
        self._outbox.put(args)

    def update(self, record_id, **kwargs):
        """Update a function evaluation status with a call to update_dict.

        Args:
            record_id: Identifier for the function evaluation
            kwargs: Named argument values
        """
        self.send('update_dict', record_id, kwargs)

    def running(self, record_id):
        """Indicate that a function evaluation is running.

        Args:
            record_id: Identifier for the function evaluation
        """
        self.send('running', record_id)

    def finish_success(self, record_id, value):
        """Indicate that a function evaluation completed successfully.

        Args:
            record_id: Identifier for the function evaluation
            value: Value returned by the feval
        """
        self.send('complete', record_id, value)

    def finish_cancel(self, record_id):
        """Indicate that a function evaluation was cancelled (at worker).

        Args:
            record_id: Identifier for the function evaluation
        """
        self.send('cancel', record_id)

    def finish_killed(self, record_id):
        """Indicate that a function evaluation was killed (controller request).

        Args:
            record_id: Identifier for the function evaluation
        """
        self.send('kill', record_id)

    def _handle_message(self):
        "Handle an incoming message and dispatch to handler."
        logger.debug("Handle incoming message")
        try:
            s = MPI.Status()
            data = comm.recv(status=s, source=0, tag=0)
            logger.debug("Incoming message received")
            mname = "on_{0}".format(data[0])
            logger.debug("Call to %s%s", mname, data[1:])
            method = getattr(self, mname)
            method(*data[1:])
        except Exception:
            logger.debug("Exception in message handler")

    def run(self):
        "Run the main loop."
        self._running = True
        while self._running:
            if comm.Iprobe():
                self._handle_message()
            elif self._eval_thread and not self._eval_thread.is_alive():
                logger.debug("Join worker eval thread")
                self._eval_thread.join()
                self._eval_thread = None
                logger.debug("Joined eval thread")
            else:
                try:
                    timeout = 0.005
                    msg = self._outbox.get(True, timeout)
                    logger.debug("MPI send to 0: %s", msg)
                    comm.send(msg, dest=0, tag=0)
                    logger.debug("MPI send completed")
                except Queue.Empty:
                    pass


class MPISimpleWorker(MPIWorker):
    """Worker that calls a Python function.

    The MPISimpleWorker does ordinary Python function evaluations.
    Requests to kill a running evaluation are simply ignored.
    """

    def __init__(self, f):
        super(MPISimpleWorker, self).__init__()
        self.f = f

    def eval(self, record_id, params):
        """Evaluate a function at a point.

        Args:
            record_id: Identifier for the function evaluation
            params: Set of parameters
        """
        logger.debug("Eval %d at %s", record_id, params)
        value = self.f(*params)
        self.finish_success(record_id, value)


class MPIProcessWorker(MPIWorker):
    """MPI worker that runs an evaluation in a subprocess

    The MPIProcessWorker is a base class for simulations that run a
    simulation in an external subprocess.  This class provides functionality
    just to allow graceful termination of the external simulations.

    Attributes:
        process: Handle for external subprocess
    """

    def __init__(self):
        super(MPIProcessWorker, self).__init__()
        self.process = None

    def kill_process(self):
        "Kill the child process"
        if self.process is not None and self.process.poll() is None:
            logger.debug("MPIProcessWorker is killing subprocess")
            self.process.terminate()

    def on_kill(self, record_id):
        self._eval_killed = True
        self.kill_process()

    def on_terminate(self):
        self.kill_process()
        super(MPIProcessWorker, self).on_terminate()



# GridCal Helm Formulation Tests - 25/04/2020
# Andre Lazaro Souza

# This file is part of GridCal.
#
# GridCal is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GridCal is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GridCal.  If not, see <http://www.gnu.org/licenses/>.

# AUTHORS: Josep Fanals Batllori and Santiago Peñate Vera
# CONTACT:  u1946589@campus.udg.edu and santiago.penate.vera@gmail.com
# thanks to Llorenç Fanals Batllori for his help at coding
import pandas as pd
import numpy as np
import numba as nb
import time
from warnings import warn
from scipy.sparse import csc_matrix, coo_matrix
from scipy.sparse import hstack as hs, vstack as vs
from scipy.sparse.linalg import spsolve, factorized


def epsilon(Sn, n, E):
    """
    Fast recursive Wynn's epsilon algorithm from:
        NONLINEAR SEQUENCE TRANSFORMATIONS FOR THE ACCELERATION OF CONVERGENCE
        AND THE SUMMATION OF DIVERGENT SERIES

        by Ernst Joachim Weniger
    Args:
        Sn: sum of coefficients
        n: order
        E: Coefficients structure copy that is modified in this algorithm

    Returns:

    """
    Zero = complex(0)
    One = complex(1)
    Tiny = np.finfo(complex).min
    Huge = np.finfo(complex).max

    E[n] = Sn

    if n == 0:
        estim = Sn
    else:
        AUX2 = Zero

        for j in range(n, 0, -1):  # range from n to 1 (both included)
            AUX1 = AUX2
            AUX2 = E[j - 1]
            DIFF = E[j] - AUX2

            if abs(DIFF) <= Tiny:
                E[j - 1] = Huge
            else:
                if DIFF == 0:
                    DIFF = Tiny
                E[j - 1] = AUX1 + One / DIFF

        if np.mod(n, 2) == 0:
            estim = E[0]
        else:
            estim = E[1]

    return estim, E


@nb.njit("(c16[:])(i8, c16[:, :], f8)", cache=True)
def pade4all(order, coeff_mat, s=1.0):
    """
    Computes the "order" Padè approximant of the coefficients at the approximation point s

    Arguments:
        coeff_mat: coefficient matrix (order, buses)
        order:  order of the series
        s: point of approximation (at 1 you get the voltage)

    Returns:
        Padè approximation at s for all the series
    """
    nbus = coeff_mat.shape[1]

    complex_type = nb.complex128

    voltages = np.zeros(nbus, dtype=complex_type)

    nn = int(order / 2)
    L = nn
    M = nn

    for d in range(nbus):

        # formation of the linear system right hand side
        rhs = coeff_mat[L + 1:L + M + 1, d]

        # formation of the coefficients matrix
        C = np.zeros((L, M), dtype=complex_type)
        for i in range(L):
            k = i + 1
            C[i, :] = coeff_mat[L - M + k:L + k, d]

        # Obtaining of the b coefficients for orders greater than 0
        b = np.zeros(rhs.shape[0] + 1, dtype=complex_type)
        x = np.linalg.solve(C, -rhs)  # bn to b1
        b[0] = 1
        b[1:] = x[::-1]

        # Obtaining of the coefficients 'a'
        a = np.zeros(L + 1, dtype=complex_type)
        a[0] = coeff_mat[0, d]
        for i in range(L):
            val = complex_type(0)
            k = i + 1
            for j in range(k + 1):
                val += coeff_mat[k - j, d] * b[j]
            a[i + 1] = val

        # evaluation of the function for the value 's'
        p = complex_type(0)
        q = complex_type(0)
        for i in range(L + 1):
            p += a[i] * s ** i
            q += b[i] * s ** i

        voltages[d] = p / q

    return voltages


@nb.njit("(c16[:])(c16[:, :], c16[:, :], i8, c16[:])", cache=True)
def sigma_function(coeff_matU, coeff_matX, order, V_slack):
    """

    :param coeff_matU: array with voltage coefficients
    :param coeff_matX: array with inverse conjugated voltage coefficients
    :param order: should be prof - 1
    :param V_slack: slack bus voltage vector. Must contain only 1 slack bus
    :return: sigma complex value
    """
    if len(V_slack) > 1:
        print('Sigma values may not be correct')
    V0 = V_slack[0]
    coeff_matU = coeff_matU / V0
    coeff_matX = coeff_matX / V0
    nbus = coeff_matU.shape[1]
    complex_type = nb.complex128
    sigmes = np.zeros(nbus, dtype=complex_type)

    if order % 2 == 0:
        M = int(order / 2) - 1
    else:
        M = int(order / 2)

    for d in range(nbus):
        a = coeff_matU[1:2 * M + 2, d]
        b = coeff_matX[0:2 * M + 1, d]
        C = np.zeros((2 * M + 1, 2 * M + 1), dtype=complex_type)

        for i in range(2 * M + 1):
            if i < M:
                C[1 + i:, i] = a[:2 * M - i]
            else:
                C[i - M:, i] = - b[:3 * M - i + 1]

        lhs = np.linalg.solve(C, -a)

        sigmes[d] = np.sum(lhs[M:]) / (np.sum(lhs[:M]) + 1)

    return sigmes


@nb.njit("(c16[:])(c16[:, :], c16[:, :], i8, i8[:])", cache=True)
def conv1(A, B, c, indices):
    """
    Performs the convolution of A* and B
    :param A: Coefficients matrix 1 (orders, buses)
    :param B: Coefficients matrix 2 (orders, buses)
    :param c: order of the coefficients
    :param indices: bus indices array
    :return: Array with the convolution for the buses given by "indices"
    """
    suma = np.zeros(len(indices), dtype=nb.complex128)
    for k in range(1, c + 1):
        for i, d in enumerate(indices):
            suma[i] += np.conj(A[k, d]) * B[c - k, d]
    return suma


@nb.njit("(c16[:])(c16[:, :], c16[:, :], i8, i8[:])", cache=True)
def conv2(A, B, c, indices):
    """
    Performs the convolution of A and B
    :param A: Coefficients matrix 1 (orders, buses)
    :param B: Coefficients matrix 2 (orders, buses)
    :param c: order of the coefficients
    :param indices: bus indices array
    :return: Array with the convolution for the buses given by "indices"
    """
    suma = np.zeros(len(indices), dtype=nb.complex128)
    for k in range(1, c):
        for i, d in enumerate(indices):
            suma[i] += A[k, d] * B[c - 1 - k, d]
    return suma


@nb.njit("(c16[:])(c16[:, :], c16[:, :], i8, i8[:])", cache=True)
def conv3(A, B, c, indices):
    """
    Performs the convolution of A and B*
    :param A: Coefficients matrix 1 (orders, buses)
    :param B: Coefficients matrix 2 (orders, buses)
    :param c: order of the coefficients
    :param indices: bus indices array
    :return: Array with the convolution for the buses given by "indices"
    """
    suma = np.zeros(len(indices), dtype=nb.complex128)
    for k in range(1, c):
        for i, d in enumerate(indices):
            suma[i] += A[k, d] * np.conj(B[c - k, d])
    return suma


def helm_coefficients_josep(Yseries, V0, S0, Ysh0, pq, pv, sl, pqpv, tolerance=1e-6, max_coeff=30, verbose=False):
    """
    Holomorphic Embedding LoadFlow Method as formulated by Josep Fanals Batllori in 2020
    
    (----Without Slack Embedding and other tweaks----) 
                                                                                 
    THis function just returns the coefficients for further usage in other routines
    :param Yseries: Admittance matrix of the series elements
    :param V0: vector of specified voltages
    :param S0: vector of specified power
    :param Ysh0: vector of shunt admittances (including the shunts of the branches)
    :param pq: list of pq nodes
    :param pv: list of pv nodes
    :param sl: list of slack nodes
    :param pqpv: sorted list of pq and pv nodes
    :param tolerance: target error (or tolerance)
    :param max_coeff: maximum number of coefficients
    :param verbose: print intermediate information
    :return: U, X, Q, iterations
    """

    npqpv = len(pqpv)
    npv = len(pv)
    nsl = len(sl)
    n = Yseries.shape[0]

    # --------------------------- PREPARING IMPLEMENTATION -------------------------------------------------------------
    U = np.zeros((max_coeff, npqpv), dtype=complex)  # voltages
    U_re = np.zeros((max_coeff, npqpv), dtype=float)  # real part of voltages
    U_im = np.zeros((max_coeff, npqpv), dtype=float)  # imaginary part of voltages
    X = np.zeros((max_coeff, npqpv), dtype=complex)  # compute X=1/conj(U)
    X_re = np.zeros((max_coeff, npqpv), dtype=float)  # real part of X
    X_im = np.zeros((max_coeff, npqpv), dtype=float)  # imaginary part of X
    Q = np.zeros((max_coeff, npqpv), dtype=complex)  # unknown reactive powers
    Vm0 = np.abs(V0)
    vec_W = Vm0 * Vm0

    if n < 2:
        return U, X, Q, 0

    if verbose:
        print('Yseries')
        print(Yseries.toarray())
        df = pd.DataFrame(data=np.c_[Ysh0.imag, S0.real, S0.imag, Vm0],
                          columns=['Ysh', 'P0', 'Q0', 'V0'])
        print(df)

    Yred = Yseries[np.ix_(pqpv, pqpv)]  # admittance matrix without slack buses
    Yslack = -Yseries[np.ix_(pqpv, sl)]  # yes, it is the negative of this
    G = np.real(Yred)  # real parts of Yij
    B = np.imag(Yred)  # imaginary parts of Yij
    vec_P = S0.real[pqpv]
    vec_Q = S0.imag[pqpv]
    Vslack = V0[sl]
    Ysh = -Ysh0[pqpv]

    # indices 0 based in the internal scheme
    nsl_counted = np.zeros(n, dtype=int)
    compt = 0
    for i in range(n):
        if i in sl:
            compt += 1
        nsl_counted[i] = compt

    pq_ = pq - nsl_counted[pq]
    pv_ = pv - nsl_counted[pv]
    #pqpv_ = np.sort(np.r_[pq_, pv_])
    pqpv_ = np.arange(0,npqpv,dtype=np.int64)

    # .......................CALCULATION OF TERMS [0] ------------------------------------------------------------------

    #if nsl > 1:
    #    U[0, :] = spsolve(Yred, Yslack.sum(axis=1))
    #else:
    #    U[0, :] = spsolve(Yred, Yslack)
    U[0, :] = spsolve(Yred, Yslack * Vslack)

    X[0, :] = 1 / np.conj(U[0, :])
    U_re[0, :] = U[0, :].real
    U_im[0, :] = U[0, :].imag
    X_re[0, :] = X[0, :].real
    X_im[0, :] = X[0, :].imag

    # .......................CALCULATION OF TERMS [1] ------------------------------------------------------------------
    valor = np.zeros(npqpv, dtype=complex)

    # get the current injections that appear due to the slack buses reduction
    #I_inj_slack = Yslack[pqpv_, :] * Vslack

    #valor[pq_] = I_inj_slack[pq_] - Yslack[pq_].sum(axis=1).A1 + (vec_P[pq_] - vec_Q[pq_] * 1j) * X[0, pq_] + U[0, pq_] * Ysh[pq_]
    #valor[pv_] = I_inj_slack[pv_] - Yslack[pv_].sum(axis=1).A1 + (vec_P[pv_]) * X[0, pv_] + U[0, pv_] * Ysh[pv_]

    valor[pq_] = (vec_P[pq_] - vec_Q[pq_] * 1j) * X[0, pq_] + U[0, pq_] * Ysh[pq_]
    valor[pv_] = (vec_P[pv_]) * X[0, pv_] + U[0, pv_] * Ysh[pv_]

    # compose the right-hand side vector
    RHS = np.r_[valor.real,
                valor.imag,
                vec_W[pv] - (U[0, pv_] * U[0, pv_]).real]

    # Form the system matrix (MAT)
    Upv = U[0, pv_]
    Xpv = X[0, pv_]
    #VRE = coo_matrix((2 * Upv.real, (np.arange(npv), pv_)), shape=(npv, npqpv)).tocsc()
    #VIM = coo_matrix((2 * Upv.imag, (np.arange(npv), pv_)), shape=(npv, npqpv)).tocsc()
    VRE = coo_matrix((Upv.real, (np.arange(npv), pv_)), shape=(npv, npqpv)).tocsc()
    VIM = coo_matrix((Upv.imag, (np.arange(npv), pv_)), shape=(npv, npqpv)).tocsc()
    XIM = coo_matrix((-Xpv.imag, (pv_, np.arange(npv))), shape=(npqpv, npv)).tocsc()
    XRE = coo_matrix((Xpv.real, (pv_, np.arange(npv))), shape=(npqpv, npv)).tocsc()
    EMPTY = csc_matrix((npv, npv))

    #MAT = vs((hs((G,  -B,   XIM)),
    #          hs((B,   G,   XRE)),
    #          hs((VRE, VIM, EMPTY))), format='csc')
    MAT = vs((hs((G,  -B,   XIM), format='csr'),
              hs((B,   G,   XRE), format='csr'),
              hs((VRE, VIM, EMPTY), format='csr')), format='csc')

    if verbose:
        print('MAT')
        print(MAT.toarray())

    # factorize (only once)
    #MAT_LU = factorized(MAT.tocsc())
    MAT_LU = factorized(MAT)

    # solve
    LHS = MAT_LU(RHS)

    # update coefficients
    U[1, :] = LHS[:npqpv] + 1j * LHS[npqpv:2 * npqpv]
    Q[0, pv_] = LHS[2 * npqpv:]
    X[1, :] = -X[0, :] * np.conj(U[1, :]) / np.conj(U[0, :])

    # .......................CALCULATION OF TERMS [>=2] ----------------------------------------------------------------
    iter_ = 1
    range_pqpv = np.arange(npqpv, dtype=np.int64)
    for c in range(2, max_coeff):  # c defines the current depth

        valor[pq_] = (vec_P[pq_] - vec_Q[pq_] * 1j) * X[c - 1, pq_] + U[c - 1, pq_] * Ysh[pq_]
        valor[pv_] = conv2(X, Q, c, pv_) * -1j + U[c - 1, pv_] * Ysh[pv_] + X[c - 1, pv_] * vec_P[pv_]

        RHS = np.r_[valor.real,
                    valor.imag,
                    -aaux_conv3(U, c, pv_)]
                    #-aux_conv3(U[:,pv_], c, npv)]
                    #-conv3(U, U, c, pv_).real]

        LHS = MAT_LU(RHS)

        # update voltage coefficients
        U[c, :] = LHS[:npqpv] + 1j * LHS[npqpv:2 * npqpv]

        # update reactive power
        Q[c - 1, pv_] = LHS[2 * npqpv:]

        # update voltage inverse coefficients
        X[c, range_pqpv] = -conv1(U, X, c, range_pqpv) / np.conj(U[0, range_pqpv])

        iter_ += 1

    return U, X, Q, iter_


def helm_josep(Ybus, Yseries, V0, S0, Ysh0, pq, pv, sl, pqpv, tolerance=1e-6, max_coeff=30, use_pade=True,
               verbose=False):
    """
    Holomorphic Embedding LoadFlow Method as formulated by Josep Fanals Batllori in 2020
    
    (----Without Slack Embedding and other tweaks----) 
    
    :param Ybus: Complete admittance matrix
    :param Yseries: Admittance matrix of the series elements
    :param V0: vector of specified voltages
    :param S0: vector of specified power
    :param Ysh0: vector of shunt admittances (including the shunts of the branches)
    :param pq: list of pq nodes
    :param pv: list of pv nodes
    :param sl: list of slack nodes
    :param pqpv: sorted list of pq and pv nodes
    :param tolerance: target error (or tolerance)
    :param max_coeff: maximum number of coefficients
    :param use_pade: Use the Padè approximation? otherwise a simple summation is done
    :param verbose: print intermediate information
    :return: V, converged, norm_f, Scalc, iter_, elapsed
    """

    start_time = time.time()

    # compute the series of coefficients
    U, X, Q, iter_ = helm_coefficients_josep(Yseries, V0, S0, Ysh0, pq, pv, sl, pqpv,
                                             tolerance=tolerance, max_coeff=max_coeff, verbose=verbose)

    # --------------------------- RESULTS COMPOSITION ------------------------------------------------------------------
    if verbose:
        print('V coefficients')
        print(U)

    # compute the final voltage vector
    V = V0.copy()
    if use_pade:
        try:
            V[pqpv] = pade4all(max_coeff - 1, U, 1)
        except:
            warn('Padè failed :(, using coefficients summation')
            V[pqpv] = U.sum(axis=0)
    else:
        V[pqpv] = U.sum(axis=0)

    # compute power mismatch
    Scalc = V * np.conj(Ybus * V)
    dP = np.abs(S0[pqpv].real - Scalc[pqpv].real)
    dQ = np.abs(S0[pq].imag - Scalc[pq].imag)
    norm_f = np.linalg.norm(np.r_[dP, dQ], np.inf)  # same as max(abs())

    # check convergence
    converged = norm_f < tolerance

    elapsed = time.time() - start_time

    return V, converged, norm_f, Scalc, iter_, elapsed


@nb.njit("(f8[:])(c16[:, :], i8, i8)", cache=True)  # Input argument types
def aux_conv3(A, c, b):                 # Function with Numba Parallelization
    """
    Performs the convolution of A and A* divided by 2 (vectorized)
    :param A: coefficients matrix 1 (orders, selected buses)
    :param b: number of selected buses
    :param c: last order of the coefficients in while loop
    :param indices: bus indices array
    :return: Array with the convolution for the buses given by "indices"
    """

    # To remember: Python indexing doesn't support float

    is_even = c % 2 == 0

    d = (c + 1) // 2

    suma_xy = np.zeros(b, dtype=nb.complex128)
    
    for m in range(1, d):
        suma_xy += A[m, :] * np.conj(A[c-m, :])

    if is_even:
        suma_xy += 0.5 * A[d, :] * np.conj(A[d, :])

    return suma_xy.real


@nb.njit("(f8[:])(c16[:, :], i8, i8[:])", cache=True)
def aaux_conv3(A, c, b):
    """
    Performs the convolution of A and A* divided by 2 (enumerate-for-loop)
    :param A: coefficients matrix 1 (orders, buses)
    :param b: bus indices array
    :param c: last order of the coefficients in while loop
    :param indices: bus indices array
    :return: Array with the convolution for the buses given by "indices"
    """

    is_even = c % 2 == 0

    d = (c + 1) // 2

    suma_xy = np.zeros(len(b), dtype=nb.complex128)
    
    for m in range(1, d):                                # for m = 1:(d-1)     # | -> RANGE FOR
        for i, k in enumerate(b):                        #   i = 0;            # | -----
            suma_xy[i] += A[m, k] * np.conj(A[c-m, k])   #   for k = indices   # | -> ENUMERATE FOR
                                                         #      i = i + 1;     # | -----
    if is_even:
        for i, k in enumerate(b):
            suma_xy[i] += 0.5 * A[d, k] * np.conj(A[d, k])

    return suma_xy.real


@nb.njit("(f8[:])(c16[:, :], c16[:, :], i8, i8[:])", cache=True)
def conv4(A, B, n, idx):
    """
    Performs the convolution of A and B where B = A* divided by 2 (enumerate-for-loop)
    :param A: Coefficients matrix 1 (orders, buses)
    :param B: Coefficients matrix 2 (orders, buses)
    :param n: last order of the coefficients in while loop
    :param idx: bus indices array
    :return: Array with the convolution for the buses given by "idx"
    """

    is_even = n % 2 == 0

    nn = (n + 1) // 2

    sum_xy = np.zeros(len(idx), dtype=nb.complex128)
    
    for m in range(1, nn):
        for i, k in enumerate(idx):
            sum_xy[i] += A[m, k] * B[n-m, k]

    if is_even:
        for i, k in enumerate(idx):
            sum_xy[i] += 0.5 * A[nn, k] * B[nn, k]

    return sum_xy.real


@nb.njit("(f8[:])(c16[:, :], c16[:, :], i8, i8[:])", cache=True)
def aux_conv4(A, B, n, idx):
    """
    Performs the convolution of A and B where B = A* divided by 2 (enumerate-for-loop)
    :param A: Coefficients matrix 1 (orders, buses)
    :param B: Coefficients matrix 2 (orders, buses)
    :param n: last order of the coefficients in while loop
    :param idx: bus indices array
    :return: Array with the convolution for the buses given by "idx"
    """
    suma = np.zeros(len(idx), dtype=nb.complex128)
    
    for m in range(1, n):
        for i, k in enumerate(idx):
            suma[i] += A[m, k] * B[n-m, k]

    return 0.5 * suma.real


@nb.njit("(f8[:])(c16[:, :], c16[:, :], i8, i8[:])", cache=True)
def conv5(A, B, n, idx):
    
    """
    Performs the convolution of A and B where B = A* (enumerate-for-loop)
    :param A: Coefficients matrix 1 (orders, buses)
    :param B: Coefficients matrix 2 (orders, buses)
    :param n: last order of the coefficients in while loop
    :param idx: bus indices array
    :return: Array with the convolution for the buses given by "idx"
    """

    is_odd = n % 2 != 0

    nn = n // 2

    sum_xy = np.zeros(len(idx), dtype=nb.complex128)
    
    for m in range(0, nn):
        for i, k in enumerate(idx):
            sum_xy[i] += A[m, k] * B[n-1-m, k]

    if is_odd:
        for i, k in enumerate(idx):
            sum_xy[i] += 0.5 * A[nn, k] * B[nn, k]

    return 2 * sum_xy.real


@nb.njit("(f8[:])(c16[:, :], c16[:, :], i8, i8[:])", cache=True)
def aux_conv5(A, B, n, idx):
    """
    Performs the convolution of A and B where B = A* (enumerate-for-loop)
    :param A: Coefficients matrix 1 (orders, buses)
    :param B: Coefficients matrix 2 (orders, buses)
    :param c: last order of the coefficients in while loop
    :param indices: bus indices array
    :return: Array with the convolution for the buses given by "indices"
    """
    suma = np.zeros(len(idx), dtype=nb.complex128)
    
    for m in range(0, n):
        for i, k in enumerate(idx):
            suma[i] += A[m, k] * B[n-1-m, k]

    return suma.real


def conv6(U, conjU, Yrr, nb, n):
    
    suma = np.zeros(nb,dtype=np.complex128)
    
    for m in range(1,n):
            suma += conjU[m,:] * ( Yrr * U[n-m,:] )
            
    return suma


def helm_coefficients_andre(Yseries, V0, S0, Ysh0, pq, pv, sl, pqpv, tolerance=1e-6, max_coeff=30, verbose=False):
    """
    Holomorphic Embedding LoadFlow Method as formulated by Josep Fanals Batllori in 2020
    
    (----Reformulation to avoid Qk[n], Wk[n] calculation----)
    
    This function just returns the coefficients for further usage in other routines
    :param Yseries: Admittance matrix of the series elements
    :param V0: vector of specified voltages (complex)
    :param S0: vector of specified power (complex)
    :param Ysh0: vector of shunt admittances (including the shunts of the branches)
    :param pq: list of pq nodes
    :param pv: list of pv nodes
    :param sl: list of slack nodes
    :param pqpv: sorted list of pq and pv nodes
    :param tolerance: target error (or tolerance)
    :param max_coeff: maximum number of coefficients
    :param verbose: print intermediate information
    :return: U, "iterations"
    """

    npqpv = len(pqpv)     # number of non-slack buses
    npq = len(pq)         # number of pq buses
    npv = len(pv)         # number of pv buses
    nsl = len(sl)         # number of slack buses
    n = Yseries.shape[0]  # number of buses; size(Yseries,n) = Yseries.shape[n-1]; size(Yseries) = Yseries.shape;

    # --------------------------- PREPARING IMPLEMENTATION -------------------------------------------------------------
    
    U = np.zeros((max_coeff, npqpv), dtype=complex, order='C')  # voltage series coefficients
    conjU = np.zeros((max_coeff, npqpv), dtype=complex, order='C')  # conjugate voltage series coefficients
    Vsp  = np.abs(V0[pqpv])  # Specified voltage magnitude for PV-Buses
    Vsp_2 = Vsp * Vsp  # Specified voltage magnitude for PV-Buses Squared

    if n < 2:
        return U, 0

    if verbose:
        print('Yseries')
        print(Yseries.toarray())
        df = pd.DataFrame(data=np.c_[Ysh0.imag,S0.real,S0.imag,np.abs(V0)],
                          columns=['Ysh', 'P0', 'Q0', 'V0'])
        print(df)

    Yrr = Yseries[np.ix_(pqpv, pqpv)]  # Y_RED-RED; Yseries(pqpv,pqpv)
    Yrs = Yseries[np.ix_(pqpv, sl)]    # Y_RED-SLK
    Vs = V0[sl]                        # V_SLK
    Isl = Yrs * Vs                     # I_SLK
    Ssp = S0[pqpv]                     # Specified apparent power (Active for PQ and PV; Reactive for PQ)
    Ysh = Ysh0[pqpv]                   # Y_RED

    # indices 0 based in the internal scheme
    nsl_counted = np.zeros(n, dtype=int)
    compt = 0
    for i in range(n):        
        if i in sl:           
            compt += 1        
        nsl_counted[i] = compt

    pq_ = pq - nsl_counted[pq]
    pv_ = pv - nsl_counted[pv]
    pqpv_ = np.arange(0,npqpv,dtype=np.int64)

    # .......................CALCULATION OF TERMS [0] ------------------------------------------------------------------

    U[0, :] = spsolve(Yrr, -Isl)

    conjU[0, :] = np.conj(U[0, :])

    # .......................LINEAR SYSTEM MATRIX (MAT) ---------------------------------------------------------------- 

    YV = Yrr.multiply(conjU[0, :].reshape(npqpv, 1))
    YV = YV.tocsc()

    # | Re{YxconjV} -Im{YxconjV} | * |LHS_ReU [PQ,PV]|  = |RHS_P [PQ,PV]|
    # | Im{YxconjV}  Re{YxconjV} |   |LHS_ImU [PQ,PV]|    |RHS_Q [PQ]   |
    # | Re{V}              Im{V} |                        |RHS_V [PV]   |
    
    ReYVp = YV.real
    ImYVp = YV.imag
    ReYVq = ReYVp[np.ix_(pq_, pqpv_)]
    ImYVq = ImYVp[np.ix_(pq_, pqpv_)]
    ReV = coo_matrix((U[0, pv_].real, (np.arange(0, npv, dtype=np.int64), pv_)), shape=(npv, npqpv)).tocsc()
    ImV = coo_matrix((U[0, pv_].imag, (np.arange(0, npv, dtype=np.int64), pv_)), shape=(npv, npqpv)).tocsc()

    # sparse matrix concatenation
    MAT = vs((hs((ReYVp, -ImYVp), format='csr'),
              hs((ImYVq, ReYVq), format='csr'),
              hs((ReV  ,   ImV), format='csr')), format='csc')

    if verbose:
        print('MAT')
        print(MAT.toarray())

    # factorize (only once)
    MAT_LU = factorized(MAT)

    # .......................CALCULATION OF TERMS [>=1] ----------------------------------------------------------------

    for c in range(1, max_coeff):  # max_coeff_order

        if c == 1:
            UconjU   = U[0, :] * conjU[0, :]
            UconjU   = UconjU.real
            valor_v  = 0.5 * (Vsp_2[pv_] - UconjU[pv_])
            valor_pq = np.conj(Ssp) - Ysh * UconjU
        else:
            conjU[c-1, :] = np.conj(U[c-1, :])
            valor_v  = - conv4(U, conjU, c, pv_)
            valor_pq = - Ysh * conv5(U, conjU, c, pqpv_) - conv6(U, conjU, Yrr, npqpv, c)

        # compose the right-hand side vector
        RHS = np.r_[valor_pq.real, valor_pq[pq_].imag, valor_v]

        # solve
        LHS = MAT_LU(RHS)

        # update voltage coefficients
        U[c, :] = LHS[0:npqpv] + 1j * LHS[npqpv:2*npqpv]

    return U, max_coeff - 1


def helm_andre(Ybus, Yseries, V0, S0, Ysh0, pq, pv, sl, pqpv, tolerance=1e-6, max_coeff=30, use_pade=True,
               verbose=False):
    """
    Holomorphic Embedding LoadFlow Method as formulated by Josep Fanals Batllori in 2020
    
    (----Reformulation to avoid Qk[n], Wk[n] calculation----)
    
    :param Ybus: Complete admittance matrix
    :param Yseries: Admittance matrix of the series elements
    :param V0: vector of specified voltages
    :param S0: vector of specified power
    :param Ysh0: vector of shunt admittances (including the shunts of the branches)
    :param pq: list of pq nodes
    :param pv: list of pv nodes
    :param sl: list of slack nodes
    :param pqpv: sorted list of pq and pv nodes
    :param tolerance: target error (or tolerance)
    :param max_coeff: maximum number of coefficients
    :param use_pade: Use the Padè approximation? otherwise a simple summation is done
    :param verbose: print intermediate information
    :return: V, converged, norm_f, Scalc, iter_, elapsed
    """

    start_time = time.time()

    # compute the series of coefficients
    U, iter_ = helm_coefficients_andre(Yseries, V0, S0, Ysh0, pq, pv, sl, pqpv,
                                       tolerance=tolerance, max_coeff=max_coeff, verbose=verbose)

    # --------------------------- RESULTS COMPOSITION ------------------------------------------------------------------
    if verbose:
        print('V coefficients')
        print(U)

    # compute the final voltage vector
    V = V0.copy()
    if use_pade:
        try:
            V[pqpv] = pade4all(max_coeff - 1, U, 1)
        except:
            warn('Padè failed :(, using coefficients summation')
            V[pqpv] = U.sum(axis=0)
    else:
        V[pqpv] = U.sum(axis=0)

    # compute power mismatch
    Scalc = V * np.conj(Ybus * V)
    dP = np.abs(S0[pqpv].real - Scalc[pqpv].real)
    dQ = np.abs(S0[pq].imag - Scalc[pq].imag)
    norm_f = np.linalg.norm(np.r_[dP, dQ], np.inf)  # same as max(abs())

    # check convergence
    converged = norm_f < tolerance

    elapsed = time.time() - start_time

    return V, converged, norm_f, Scalc, iter_, elapsed


if __name__ == '__main__':
    from GridCal.Engine import FileOpen
    import pandas as pd

    np.set_printoptions(linewidth=2000, suppress=True)
    pd.set_option('display.max_rows', 500)
    pd.set_option('display.max_columns', 500)
    pd.set_option('display.width', 1000)

    # fname = '/home/santi/Documentos/GitHub/GridCal/Grids_and_profiles/grids/IEEE39_1W.gridcal'
    # fname = '/home/santi/Documentos/GitHub/GridCal/Grids_and_profiles/grids/IEEE 14.xlsx'
    # fname = '/home/santi/Documentos/GitHub/GridCal/Grids_and_profiles/grids/lynn5buspv.xlsx'
    # fname = '/home/santi/Documentos/GitHub/GridCal/Grids_and_profiles/grids/IEEE 118.xlsx'
    # fname = '/home/santi/Documentos/GitHub/GridCal/Grids_and_profiles/grids/1354 Pegase.xlsx'
    # fname = '/home/santi/Documentos/GitHub/GridCal/Grids_and_profiles/grids/Pegase 2869.xlsx'
    #fname = 'helm_data1.gridcal'
    fname = "/home/santi/Documentos/GitHub/GridCal/Grids_and_profiles/grids/Iwamoto's 11 Bus.gridcal"
    # fname = 'C:/Users/Lazaro/Downloads/Pegase 2869.xlsx'
    #fname = 'C:/Users/Lazaro/Downloads/IEEE 14_nopv.xlsx'
    #fname = 'C:/Users/Lazaro/Downloads/IEEE 14.xlsx'

    grid = FileOpen(fname).open()

    nc = grid.compile_snapshot()
    inputs = nc.compute()[0]  # pick the first island

    V, converged_, error, Scalc_, iter_, elapsed_ = helm_josep(Ybus=inputs.Ybus,
                                                               Yseries=inputs.Yseries,
                                                               V0=inputs.Vbus,
                                                               S0=inputs.Sbus,
                                                               Ysh0=inputs.Ysh,
                                                               pq=inputs.pq,
                                                               pv=inputs.pv,
                                                               sl=inputs.ref,
                                                               pqpv=inputs.pqpv,
                                                               tolerance=1e-6,
                                                               max_coeff=30,
                                                               use_pade=True,
                                                               verbose=False)
    Vm = np.abs(V)
    Va = np.angle(V)
    dP = np.abs(inputs.Sbus.real - Scalc_.real)
    dP[inputs.ref] = np.nan
    dQ = np.abs(inputs.Sbus.imag - Scalc_.imag)
    dQ[inputs.pv] = np.nan
    dQ[inputs.ref] = np.nan
    df = pd.DataFrame(data=np.c_[inputs.types, Vm, Va, np.abs(inputs.Vbus), dP, dQ],
                      columns=['Types', 'Vm', 'Va', 'Vset', 'P mismatch', 'Q mismatch'])
    # print(df)
    print('Josep (implicit slack):')
    print('Error', error)
    print('Elapsed', elapsed_)

    V, converged_, error, Scalc_, iter_, elapsed_ = helm_andre(Ybus=inputs.Ybus,
                                                               Yseries=inputs.Yseries,
                                                               V0=inputs.Vbus,
                                                               S0=inputs.Sbus,
                                                               Ysh0=inputs.Ysh,
                                                               pq=inputs.pq,
                                                               pv=inputs.pv,
                                                               sl=inputs.ref,
                                                               pqpv=inputs.pqpv,
                                                               tolerance=1e-6,
                                                               max_coeff=30,
                                                               use_pade=True,
                                                               verbose=False)
    Vm = np.abs(V)
    Va = np.angle(V)
    dP = np.abs(inputs.Sbus.real - Scalc_.real)
    dP[inputs.ref] = np.nan
    dQ = np.abs(inputs.Sbus.imag - Scalc_.imag)
    dQ[inputs.pv] = np.nan
    dQ[inputs.ref] = np.nan
    df = pd.DataFrame(data=np.c_[inputs.types, Vm, Va, np.abs(inputs.Vbus), dP, dQ],
                      columns=['Types', 'Vm', 'Va', 'Vset', 'P mismatch', 'Q mismatch'])
    # print(df)
    print('Andre (explicit slack):')
    print('Error', error)
    print('Elapsed', elapsed_)



from research.three_phase.Engine.Devices.branch import *
from research.three_phase.Engine.Devices.bus import *


class LineTypeSeq:

    def __init__(self, name, Z_SEQ, Ysh_SEQ):
        """
        3-phase branch defined from the positive sequence elements
        :param name: name of the line type object
        :param Z_SEQ: vector with the sequence impedances in Ohm/km
        :param Ysh_SEQ: vector with the shunt sequence impedances in Siemens/km
        """
        self.name = name

        self.number_of_phases = 3

        self.z_SEQ = Z_SEQ

        self.ysh_SEQ = Ysh_SEQ

    def getABCD(self, line_length, y_base):
        """
        get the line ABCD parameters

        | If |   | A  B |   | Vf |
        |    | = |      | * |    |
        | It |   | C  D |   | Vt |

        :param line_length: line length in km
        :param y_base: base admittance in Siemens
        """
        y = seq_to_abc(1.0 / (self.z_SEQ * line_length))
        y_sh = seq_to_abc(self.ysh_SEQ * line_length)

        As = y
        Ds = y
        A = y + y_sh / 2.0
        B = - y
        C = - y
        D = y + y_sh / 2.0

        return A / y_base, B / y_base, C / y_base, D / y_base, As / y_base, Ds / y_base


class LineTypeABC:

    def __init__(self, name, phases, zABC, ysh_ABC):
        """
        Rectangular full 3-phase branch
        :param name: Name of the line type
        :param phases: Number of phases
        :param zABC: phases x phases impedance matrix in Ohm/km
        :param ysh_ABC: phases x phases shunt admittance matrix in Siemens/km
        """
        self.name = name

        self.number_of_phases = phases

        self.zABC = zABC

        self.ysh_ABC = ysh_ABC

    def getABCD(self, line_length, y_base):
        """
        get the line ABCD parameters

        | If |   | A  B |   | Vf |
        |    | = |      | * |    |
        | It |   | C  D |   | Vt |

        :param line_length: line length in km
        :param y_base: Base admittance in Siemens
        """
        Yseries = np.linalg.inv(self.zABC * line_length)

        As = Yseries
        Ds = Yseries
        A = Yseries #+ self.ysh_ABC * line_length / 2.0
        B = - Yseries
        C = - Yseries
        D = Yseries #+ self.ysh_ABC * line_length / 2.0

        return A / y_base, B / y_base, C / y_base, D / y_base, As / y_base, Ds / y_base


class Line(Branch):

    def __init__(self, name, line_type, bus_from: Bus, bus_to: Bus,
                 conn_from=Phases.ABC, conn_to=Phases.ABC, length=1.0, rating=1e-6):
        """
        n_phase line
        :param name: name of the line
        :param bus_from: bus from object
        :param bus_to: bus to object
        :param conn_from: vector of connection in the bus from i.e. [0, 1, 2]
        :param conn_to: vector of connection in the bus to, i.e. [0, 1, 2]
        :param line_type: line type
        :param length: line length in km
        :param rating: transformer rating in MVA
        """
        self.name = name

        self.f = bus_from

        self.t = bus_to

        self.rating = rating

        self.line_type = line_type

        self.number_of_phases = line_type.number_of_phases

        self.length = length

        self.phases_from = np.array(conn_from)

        self.phases_to = np.array(conn_to)

        # check connection compatibility
        if len(self.phases_from) != len(self.phases_to):
            raise Exception('Wrong phases')

        if len(self.phases_from) != self.line_type.number_of_phases:
            raise Exception('The number of phases of the line type does not match the specified connection phases')

        if (self.f.Vnom / self.t.Vnom) > 1.05 or (self.t.Vnom / self.f.Vnom) > 1.05:
            raise Exception('The line ' + self.name +
                            ' is connected between buses which nominal voltage differ more than a 5%')

    def get_ABCD(self, Sbase):
        """
        get the ABCD parameters

        | If |   | A  B |   | Vf |
        |    | = |      | * |    |
        | It |   | C  D |   | Vt |
        :param Sbase: Circuit base power in MVA
        """

        # compute the base admittance
        y_base = Sbase / (self.f.Vnom * self.f.Vnom)

        return self.line_type.getABCD(self.length, y_base)

    def __str__(self):
        return self.name



# This file is part of GridCal.
#
# GridCal is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GridCal is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GridCal.  If not, see <http://www.gnu.org/licenses/>.
from numpy import complex, zeros, power

import multiprocessing
from PySide2.QtCore import QThread, Signal

from GridCal.Engine.basic_structures import Logger
from GridCal.Engine.Simulations.PowerFlow.power_flow_results import PowerFlowResults
from GridCal.Engine.Simulations.Stochastic.monte_carlo_results import MonteCarloResults
from GridCal.Engine.Simulations.Stochastic.monte_carlo_input import MonteCarloInput
from GridCal.Engine.Core.time_series_pf_data import TimeCircuit
from GridCal.Engine.Core.multi_circuit import MultiCircuit
from GridCal.Engine.basic_structures import CDF
from GridCal.Engine.Simulations.PowerFlow.power_flow_worker import PowerFlowOptions, single_island_pf, \
                                                                    power_flow_worker_args, power_flow_post_process

from GridCal.Engine.Core.time_series_pf_data import compile_time_circuit, BranchImpedanceMode

########################################################################################################################
# Monte Carlo classes
########################################################################################################################


def make_monte_carlo_input(numerical_input_island: TimeCircuit):
    """
    Generate a monte carlo input instance
    :param numerical_input_island:
    :return:
    """
    n = numerical_input_island.nbus
    Scdf = [None] * n
    Icdf = [None] * n
    Ycdf = [None] * n

    for i in range(n):
        Scdf[i] = CDF(numerical_input_island.Sbus[i, :])
        Icdf[i] = CDF(numerical_input_island.Ibus[i, :])
        Ycdf[i] = CDF(numerical_input_island.Yshunt_from_devices[i, :])

    return MonteCarloInput(n, Scdf, Icdf, Ycdf)


class MonteCarlo(QThread):
    progress_signal = Signal(float)
    progress_text = Signal(str)
    done_signal = Signal()
    name = 'Monte Carlo'

    def __init__(self, grid: MultiCircuit, options: PowerFlowOptions, mc_tol=1e-3, batch_size=100, max_mc_iter=10000,
                 opf_time_series_results=None):
        """
        Monte Carlo simulation constructor
        :param grid: MultiGrid instance
        :param options: Power flow options
        :param mc_tol: monte carlo std.dev tolerance
        :param batch_size: size of the batch
        :param max_mc_iter: maximum monte carlo iterations in case of not reach the precission
        """
        QThread.__init__(self)

        self.circuit = grid

        self.options = options

        self.opf_time_series_results = opf_time_series_results

        self.mc_tol = mc_tol

        self.batch_size = batch_size
        self.max_mc_iter = max_mc_iter

        self.results = None

        self.logger = Logger()

        self.pool = None

        self.returned_results = list()

        self.__cancel__ = False

    def get_steps(self):
        """
        Get time steps list of strings
        """
        p = self.results.points_number
        return ['point:' + str(l) for l in range(p)]

    def update_progress_mt(self, res):
        """
        """
        t, _ = res
        progress = (t + 1) / self.max_mc_iter * 100
        self.progress_signal.emit(progress)
        self.returned_results.append(res)

    def run_multi_thread(self):
        """
        Run the monte carlo simulation
        @return: MonteCarloResults instance
        """

        self.__cancel__ = False

        # initialize the grid time series results
        # we will append the island results with another function
        # self.circuit.time_series_results = TimeSeriesResults(0, 0, [])
        self.pool = multiprocessing.Pool()
        it = 0
        variance_sum = 0.0
        std_dev_progress = 0
        v_variance = 0

        n = len(self.circuit.buses)
        m = self.circuit.get_branch_number()

        mc_results = MonteCarloResults(n, m, name='Monte Carlo')
        avg_res = PowerFlowResults()
        avg_res.initialize(n, m)

        # compile circuits
        numerical_circuit = self.circuit.compile_time_series()

        # perform the topological computation
        calc_inputs_dict = numerical_circuit.compute(branch_tolerance_mode=self.options.branch_impedance_tolerance_mode,
                                                     ignore_single_node_islands=self.options.ignore_single_node_islands)

        mc_results.bus_types = numerical_circuit.bus_types

        v_sum = zeros(n, dtype=complex)

        self.progress_signal.emit(0.0)

        while (std_dev_progress < 100.0) and (it < self.max_mc_iter) and not self.__cancel__:

            self.progress_text.emit('Running Monte Carlo: Variance: ' + str(v_variance))

            mc_results = MonteCarloResults(n, m, self.batch_size, name='Monte Carlo')

            # for each partition of the profiles...
            for t_key, calc_inputs in calc_inputs_dict.items():

                # For every island, run the time series
                for island_index, numerical_island in enumerate(calc_inputs):

                    # set the time series as sampled
                    monte_carlo_input = make_monte_carlo_input(numerical_island)
                    mc_time_series = monte_carlo_input(self.batch_size, use_latin_hypercube=False)
                    Vbus = numerical_island.Vbus
                    branch_rates = numerical_island.branch_rates

                    # short cut the indices
                    b_idx = numerical_island.original_bus_idx
                    br_idx = numerical_island.original_branch_idx

                    self.returned_results = list()

                    t = 0
                    while t < self.batch_size and not self.__cancel__:

                        Ysh, Ibus, Sbus = mc_time_series.get_at(t)

                        args = (t, self.options, numerical_island, Vbus, Sbus, Ibus, branch_rates)

                        self.pool.apply_async(power_flow_worker_args, (args,), callback=self.update_progress_mt)

                    # wait for all jobs to complete
                    self.pool.close()
                    self.pool.join()

                    # collect results
                    self.progress_text.emit('Collecting batch results...')
                    for t, res in self.returned_results:
                        # store circuit results at the time index 't'
                        mc_results.S_points[t, numerical_island.original_bus_idx] = res.Sbus
                        mc_results.V_points[t, numerical_island.original_bus_idx] = res.voltage
                        mc_results.Sbr_points[t, numerical_island.original_branch_idx] = res.Sf
                        mc_results.loading_points[t, numerical_island.original_branch_idx] = res.loading
                        mc_results.losses_points[t, numerical_island.original_branch_idx] = res.losses

                    # compile MC results
                    self.progress_text.emit('Compiling results...')
                    mc_results.compile()

                    # compute the island branch results
                    island_avg_res = numerical_island.compute_branch_results(mc_results.voltage[b_idx])

                    # apply the island averaged results
                    avg_res.apply_from_island(island_avg_res, b_idx=b_idx, br_idx=br_idx)

            # Compute the Monte Carlo values
            it += self.batch_size
            mc_results.append_batch(mc_results)
            v_sum += mc_results.get_voltage_sum()
            v_avg = v_sum / it
            v_variance = abs((power(mc_results.V_points - v_avg, 2.0) / (it - 1)).min())

            # progress
            variance_sum += v_variance
            err = variance_sum / it
            if err == 0:
                err = 1e-200  # to avoid division by zeros
            mc_results.error_series.append(err)

            # emmit the progress signal
            std_dev_progress = 100 * self.mc_tol / err
            if std_dev_progress > 100:
                std_dev_progress = 100
            self.progress_signal.emit(max((std_dev_progress, it / self.max_mc_iter * 100)))

        # compute the averaged branch magnitudes
        mc_results.sbranch = avg_res.Sf
        mc_results.losses = avg_res.losses

        # print('V mc: ', mc_results.voltage)

        # send the finnish signal
        self.progress_signal.emit(0.0)
        self.progress_text.emit('Done!')
        self.done_signal.emit()

        return mc_results

    def run_single_thread(self):
        """
        Run the monte carlo simulation
        @return:
        """

        self.__cancel__ = False

        # initialize the grid time series results
        # we will append the island results with another function
        # self.circuit.time_series_results = TimeSeriesResults(0, 0, [])
        # Sbase = self.circuit.Sbase

        it = 0
        variance_sum = 0.0
        std_dev_progress = 0
        v_variance = 0

        # n = len(self.circuit.buses)
        # m = self.circuit.get_branch_number()
        #
        # # compile circuits
        # numerical_circuit = self.circuit.compile_time_series()
        #
        # # perform the topological computation
        # calc_inputs_dict = numerical_circuit.compute(branch_tolerance_mode=self.options.branch_impedance_tolerance_mode,
        #                                              ignore_single_node_islands=self.options.ignore_single_node_islands)
        #
        # mc_results = MonteCarloResults(n, m, name='Monte Carlo')
        # avg_res = PowerFlowResults()
        # avg_res.initialize(n, m)

        # compile the multi-circuit
        numerical_circuit = compile_time_circuit(circuit=self.circuit,
                                                 apply_temperature=False,
                                                 branch_tolerance_mode=BranchImpedanceMode.Specified,
                                                 opf_results=self.opf_time_series_results)

        # do the topological computation
        calculation_inputs = numerical_circuit.split_into_islands(ignore_single_node_islands=self.options.ignore_single_node_islands)

        mc_results_master = MonteCarloResults(n=numerical_circuit.nbus,
                                              m=numerical_circuit.nbr,
                                              p=self.max_mc_iter,
                                              bus_names=numerical_circuit.bus_names,
                                              branch_names=numerical_circuit.branch_names,
                                              bus_types=numerical_circuit.bus_types,
                                              name='Monte Carlo')

        avg_res = PowerFlowResults(n=numerical_circuit.nbus,
                                   m=numerical_circuit.nbr,
                                   n_tr=numerical_circuit.ntr,
                                   n_hvdc=numerical_circuit.nhvdc,
                                   bus_names=numerical_circuit.bus_names,
                                   branch_names=numerical_circuit.branch_names,
                                   transformer_names=numerical_circuit.tr_names,
                                   hvdc_names=numerical_circuit.hvdc_names,
                                   bus_types=numerical_circuit.bus_types)

        n = numerical_circuit.nbus
        m = numerical_circuit.nbr
        v_sum = zeros(n, dtype=complex)

        self.progress_signal.emit(0.0)

        while (std_dev_progress < 100.0) and (it < self.max_mc_iter) and not self.__cancel__:

            self.progress_text.emit('Running Monte Carlo: Variance: ' + str(v_variance))

            batch_results = MonteCarloResults(n=numerical_circuit.nbus,
                                              m=numerical_circuit.nbr,
                                              p=self.max_mc_iter,
                                              bus_names=numerical_circuit.bus_names,
                                              branch_names=numerical_circuit.branch_names,
                                              bus_types=numerical_circuit.bus_types,
                                              name='Monte Carlo')

            # For every island, run the time series
            for island_index, numerical_island in enumerate(calculation_inputs):

                # short cut the indices
                bus_idx = numerical_island.original_bus_idx
                br_idx = numerical_island.original_branch_idx

                # set the time series as sampled
                monte_carlo_input = make_monte_carlo_input(numerical_island)
                mc_time_series = monte_carlo_input(self.batch_size, use_latin_hypercube=False)
                Vbus = numerical_island.Vbus[:, 0]

                # run the time series
                for t in range(self.batch_size):
                    # set the power values
                    Y, I, S = mc_time_series.get_at(t)

                    res = single_island_pf(circuit=numerical_island,
                                           Vbus=Vbus,
                                           Sbus=S,
                                           Ibus=I,
                                           branch_rates=numerical_island.branch_rates,
                                           options=self.options,
                                           logger=self.logger)

                    batch_results.S_points[t, bus_idx] = res.Sbus
                    batch_results.V_points[t, bus_idx] = res.voltage
                    batch_results.Sbr_points[t, br_idx] = res.Sf
                    batch_results.loading_points[t, br_idx] = res.loading
                    batch_results.losses_points[t, br_idx] = res.losses

                self.progress_text.emit('Compiling results...')
                batch_results.compile()

                # compute the island branch results
                Sfb, Stb, If, It, Vbranch, loading, \
                losses, flow_direction, Sbus = power_flow_post_process(numerical_island,
                                                                       Sbus=batch_results.S_points.mean(axis=0)[bus_idx],
                                                                       V=batch_results.V_points.mean(axis=0)[bus_idx],
                                                                       branch_rates=numerical_island.branch_rates)

                # apply the island averaged results
                avg_res.Sbus[bus_idx] = Sbus
                avg_res.voltage[bus_idx] = batch_results.voltage[bus_idx]
                avg_res.Sf[br_idx] = Sfb
                avg_res.St[br_idx] = Stb
                avg_res.If[br_idx] = If
                avg_res.It[br_idx] = It
                avg_res.Vbranch[br_idx] = Vbranch
                avg_res.loading[br_idx] = loading
                avg_res.losses[br_idx] = losses
                avg_res.flow_direction[br_idx] = flow_direction

            # Compute the Monte Carlo values
            it += self.batch_size
            mc_results_master.append_batch(batch_results)
            v_sum += mc_results_master.get_voltage_sum()
            v_avg = v_sum / it
            v_variance = abs((power(mc_results_master.V_points - v_avg, 2.0) / (it - 1)).min())

            # progress
            variance_sum += v_variance
            err = variance_sum / it
            if err == 0:
                err = 1e-200  # to avoid division by zeros
            mc_results_master.error_series.append(err)

            # emmit the progress signal
            std_dev_progress = 100 * self.mc_tol / err
            if std_dev_progress > 100:
                std_dev_progress = 100
            self.progress_signal.emit(max((std_dev_progress, it / self.max_mc_iter * 100)))

        # compile results
        mc_results_master.bus_types = numerical_circuit.bus_types

        # send the finnish signal
        self.progress_signal.emit(0.0)
        self.progress_text.emit('Done!')
        self.done_signal.emit()

        return mc_results_master

    def run(self):
        """
        Run the monte carlo simulation
        @return:
        """
        # print('LHS run')
        self.__cancel__ = False

        if self.options.multi_thread:
            self.results = self.run_multi_thread()
        else:
            self.results = self.run_single_thread()

        # send the finnish signal
        self.progress_signal.emit(0.0)
        self.progress_text.emit('Done!')
        self.done_signal.emit()

    def cancel(self):
        """
        Cancel the simulation
        :return:
        """
        self.__cancel__ = True
        self.progress_signal.emit(0.0)
        self.progress_text.emit('Cancelled')
        self.done_signal.emit()

